# canonmap/services/db_mysql/helpers/datetime_formats.py

import dateutil.parser


DATETIME_TYPES = {"date", "datetime", "timestamp", "time", "year"}

def infer_date_format(samples):
    # Very basic inference, can be expanded
    # Try to use strftime format codes as sample
    formats = set()
    for val in samples:
        try:
            dt = dateutil.parser.parse(str(val))
            # Choose a format based on type
            if len(str(val)) == 4 and str(val).isdigit():
                fmt = "%Y"
            elif "-" in str(val):
                if ":" in str(val):
                    fmt = "%Y-%m-%d %H:%M:%S"
                else:
                    fmt = "%Y-%m-%d"
            else:
                fmt = "unknown"
            formats.add(fmt)
        except Exception:
            continue
    return list(formats) or ["unknown"]


from .core import CanonMap
from .services.db_mysql.config import CanonMapMySQLConfig
from .services.db_mysql.core import CanonMapMySQLClient
from .services.db_mysql.schemas import EntityMappingRequest, EntityMappingResponse, SingleMappedEntity, TableField, FieldTransform, FieldTransformType

# Cloud SQL helper functions
def create_cloud_sql_client(project_id=None, region=None, instance_name=None, database_name=None, user=None, password=None):
    """
    Create a CanonMapMySQLClient configured for Google Cloud SQL.
    
    This function avoids circular imports by importing locally.
    """
    from .services.db_mysql.helpers.cloud_sql import create_cloud_sql_client as _create_cloud_sql_client
    return _create_cloud_sql_client(project_id, region, instance_name, database_name, user, password)

def create_client_from_env():
    """
    Create a CanonMapMySQLClient using environment variables.
    
    This function avoids circular imports by importing locally.
    """
    from .services.db_mysql.helpers.cloud_sql import create_client_from_env as _create_client_from_env
    return _create_client_from_env()

__all__ = [
    "CanonMap",
    "CanonMapMySQLConfig",
    "CanonMapMySQLClient",
    "EntityMappingRequest",
    "EntityMappingResponse",
    "SingleMappedEntity",
    "TableField",
    "FieldTransform",
    "FieldTransformType",
    "create_cloud_sql_client",
    "create_client_from_env",
]


# canonmap/services/db_mysql/core.py

from canonmap.services.db_mysql.adapters.connection import ConnectionManager
from canonmap.services.db_mysql.managers.database import DatabaseManager
from canonmap.services.db_mysql.managers.user import UserManager
from canonmap.services.db_mysql.managers.table import TableManager
from canonmap.services.db_mysql.config import CanonMapMySQLConfig
from canonmap.services.db_mysql.managers.matcher import MatcherManager
from canonmap.services.db_mysql.managers.csv import CSVManager

class CanonMapMySQLClient:
    def __init__(self, config: CanonMapMySQLConfig):
        self.connection_manager = ConnectionManager(config)
        self.database_manager = DatabaseManager(self.connection_manager)
        self.user_manager = UserManager(self.connection_manager)
        self.table_manager = TableManager(self.connection_manager)
        self.matcher_manager = MatcherManager(self.connection_manager)
        self.csv_manager = CSVManager(self.connection_manager)


# canonmap/services/db_mysql/config.py

import os
from dataclasses import dataclass
from typing import Optional

@dataclass
class CanonMapMySQLConfig:
    host: str
    user: str
    password: str
    port: int = 3306
    database: Optional[str] = None
    unix_socket: Optional[str] = None
    cloud_sql_instance: Optional[str] = None

    @classmethod
    def from_environment(cls, prefix: str = "DB_") -> "CanonMapMySQLConfig":
        """
        Create a CanonMapMySQLConfig from environment variables.
        
        For Cloud SQL on Cloud Run, set these environment variables:
        - DB_USER: MySQL username
        - DB_PASSWORD: MySQL password  
        - DB_NAME: Database name
        - DB_UNIX_SOCKET: Unix socket path (e.g., /cloudsql/PROJECT:REGION:INSTANCE)
        
        For regular MySQL connections:
        - DB_HOST: MySQL host
        - DB_PORT: MySQL port (default: 3306)
        """
        # Check for Cloud SQL Unix socket first
        unix_socket = os.getenv(f"{prefix}UNIX_SOCKET")
        
        if unix_socket:
            # Cloud SQL via Unix socket
            return cls(
                host="localhost",  # Not used with Unix socket
                user=os.getenv(f"{prefix}USER", ""),
                password=os.getenv(f"{prefix}PASSWORD", ""),
                database=os.getenv(f"{prefix}NAME"),
                unix_socket=unix_socket,
            )
        else:
            # Regular TCP connection
            return cls(
                host=os.getenv(f"{prefix}HOST", "localhost"),
                user=os.getenv(f"{prefix}USER", ""),
                password=os.getenv(f"{prefix}PASSWORD", ""),
                port=int(os.getenv(f"{prefix}PORT", "3306")),
                database=os.getenv(f"{prefix}NAME"),
            ) 


# canonmap/services/db_mysql/managers/database.py

import logging
import pickle
import os
import json
import re
from typing import List, Optional, Dict, Any, Union
from collections import defaultdict

import mysql.connector

from canonmap.services.db_mysql.adapters.connection import ConnectionManager
from canonmap.services.db_mysql.adapters.cursor import get_cursor
from canonmap.services.db_mysql.schemas import TableField
from canonmap.services.db_mysql.helpers.datetime_formats import DATETIME_TYPES, infer_date_format

logger = logging.getLogger(__name__)

class DatabaseManager:
    def __init__(
        self, 
        connection_manager: ConnectionManager,
    ):
        self.connection_manager = connection_manager

    def connect_to_database(
        self, 
        db_name: str, 
        create_if_not_exists: bool = True,
    ) -> None:
        conn = self.connection_manager.connect()
        
        with get_cursor(conn) as cursor:
            cursor.execute(
                "SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA "
                "WHERE SCHEMA_NAME = %s",
                (db_name,)
            )
            already_exists = cursor.fetchone() is not None
            
            if already_exists:
                logger.info(f"Database '{db_name}' already exists")
            elif create_if_not_exists:
                logger.info(f"Database '{db_name}' does not exist, creating it...")
                cursor.execute(f"CREATE DATABASE `{db_name}`")
                conn.commit()
                logger.info(f"Database '{db_name}' created successfully")
            else:
                raise RuntimeError(f"Database '{db_name}' does not exist and create_if_not_exists=False")
        
        self.connection_manager.close()
        self.connection_manager.config.database = db_name
        
        config = {
            "host": self.connection_manager.config.host,
            "user": self.connection_manager.config.user,
            "password": self.connection_manager.config.password,
            "port": self.connection_manager.config.port,
            "database": db_name,
        }
        
        try:
            self.connection_manager.conn = mysql.connector.connect(**config)
            logger.info(f"Connected to database '{db_name}'")
        except mysql.connector.Error as e:
            logger.error(f"Could not connect to database '{db_name}': {e}")
            raise RuntimeError(f"Could not connect to database '{db_name}': {e}") from e

    def create_database(
        self, 
        db_name: str,
    ) -> None:
        conn = self.connection_manager.connect()
        
        with get_cursor(conn) as cursor:
            cursor.execute(
                "SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA "
                "WHERE SCHEMA_NAME = %s",
                (db_name,)
            )
            already_exists = cursor.fetchone() is not None
            
            if already_exists:
                logger.info(f"Database '{db_name}' already exists")
            else:
                logger.info(f"Database '{db_name}' does not exist, creating it...")
                cursor.execute(f"CREATE DATABASE `{db_name}`")
                conn.commit()
                logger.info(f"Database '{db_name}' created successfully")
        
        self.connection_manager.config.database = db_name
        self.connection_manager.close()
        self.connection_manager.connect()

    def delete_database(
        self, 
        database_name: str, 
        skip_confirm: bool = False,
    ) -> None:
        clean_name = database_name.strip()
        if not skip_confirm:
            resp1 = input(
                f"Are you sure you want to delete database '{clean_name}'? "
                "This will remove the entire database and all its tables. [y/N]: "
            )
            if resp1.strip().lower() not in ('y', 'yes'):
                logger.info(
                    f"Deletion of database '{clean_name}' cancelled at first prompt."
                )
                return
            resp2 = input(
                f"Type 'DELETE' to permanently delete database '{clean_name}': "
            )
            if resp2.strip() != 'DELETE':
                logger.info(
                    f"Deletion of database '{clean_name}' cancelled at second prompt."
                )
                return

        conn = self.connection_manager.connect(create_database=False)
        with get_cursor(conn) as cursor:
            cursor.execute(
                "SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA "
                "WHERE SCHEMA_NAME = %s",
                (clean_name,)
            )
            if not cursor.fetchone():
                logger.warning(f"Database '{clean_name}' does not exist; nothing to delete.")
                return

        with get_cursor(conn) as cursor:
            cursor.execute(f"DROP DATABASE `{clean_name}`;")
        conn.commit()
        logger.info(f"Database '{clean_name}' deleted successfully.")

        if self.connection_manager.config.database == clean_name:
            self.connection_manager.config.database = None




    def generate_schema(
        self,
        schema_name: str,
        fields_to_include: Optional[List[TableField]] = None,
        fields_to_exclude: Optional[List[TableField]] = None,
        tables_to_include: Optional[List[str]] = None,
        num_examples: int = 10,
        include_helper_fields: bool = False,
        save_dir: str = ".",
        save_json_version: Optional[str] = None,
    ) -> str:
        conn = self.connection_manager.connect()
        schema = defaultdict(dict)

        # fetch table/column/type info
        with get_cursor(conn) as cursor:
            if tables_to_include:
                placeholders = ', '.join(['%s'] * len(tables_to_include))
                cursor.execute(
                    f"SELECT TABLE_NAME, COLUMN_NAME, DATA_TYPE "
                    f"FROM INFORMATION_SCHEMA.COLUMNS "
                    f"WHERE TABLE_SCHEMA=%s AND TABLE_NAME IN ({placeholders})",
                    (self.connection_manager.config.database,) + tuple(tables_to_include)
                )
            else:
                cursor.execute(
                    "SELECT TABLE_NAME, COLUMN_NAME, DATA_TYPE "
                    "FROM INFORMATION_SCHEMA.COLUMNS "
                    "WHERE TABLE_SCHEMA=%s",
                    (self.connection_manager.config.database,)
                )
            columns = cursor.fetchall()

        # filter include/exclude sets
        include_set = set((f.table_name, f.field_name) for f in fields_to_include or [])
        exclude_set = set((f.table_name, f.field_name) for f in fields_to_exclude or [])

        filtered_columns = []
        for table, col, typ in columns:
            if not include_helper_fields and col.startswith("__") and col.endswith("__"):
                continue

            if fields_to_include:
                if (table, col) not in include_set:
                    continue
            elif fields_to_exclude:
                if (table, col) in exclude_set:
                    continue

            filtered_columns.append((table, col, typ))

        # sample data for each field
        with get_cursor(conn) as cursor:
            for table, col, typ in filtered_columns:
                samples: List[Any] = []

                # stratified PK-range sampling: 3/4/3 from three equal buckets
                try:
                    cursor.execute(
                        "SELECT COLUMN_NAME, DATA_TYPE FROM INFORMATION_SCHEMA.COLUMNS "
                        "WHERE TABLE_SCHEMA=%s AND TABLE_NAME=%s AND COLUMN_KEY='PRI'",
                        (self.connection_manager.config.database, table)
                    )
                    pk_info = cursor.fetchone()
                    if not pk_info or pk_info[1].upper() not in ("INT","BIGINT","SMALLINT","MEDIUMINT","TINYINT"):
                        raise ValueError("No suitable integer PK")

                    pk_col = pk_info[0]
                    cursor.execute(f"SELECT MIN(`{pk_col}`), MAX(`{pk_col}`) FROM `{table}`")
                    min_id, max_id = cursor.fetchone()
                    import random

                    total = max_id - min_id + 1
                    size = total // 3
                    buckets = [
                        (min_id,           min_id + size - 1, 3),
                        (min_id + size,    min_id + 2*size - 1, 4),
                        (min_id + 2*size,  max_id,             3),
                    ]

                    for start, end, want in buckets:
                        got = []
                        trials = 0
                        while len(got) < want and trials < want * 10:
                            trials += 1
                            rand_id = random.randint(start, end)
                            cursor.execute(
                                f"SELECT `{col}` FROM `{table}` "
                                f"WHERE `{pk_col}` >= %s AND `{col}` IS NOT NULL LIMIT 1",
                                (rand_id,)
                            )
                            row = cursor.fetchone()
                            if row and row[0] not in got:
                                got.append(row[0])
                        samples.extend(got)
                        logger.debug(
                            f"Bucket {start}–{end}: wanted {want}, got {len(got)} after {trials} trials"
                        )
                except Exception:
                    # fallback reservoir sampling on first 10000 distinct values
                    cursor.execute(
                        f"SELECT DISTINCT `{col}` FROM `{table}` "
                        "WHERE `{col}` IS NOT NULL LIMIT 10000"
                    )
                    import random
                    reservoir: List[Any] = []
                    for idx, row in enumerate(cursor):
                        val = row[0]
                        if idx < num_examples:
                            reservoir.append(val)
                        else:
                            j = random.randint(0, idx)
                            if j < num_examples:
                                reservoir[j] = val
                    samples = reservoir
                    logger.debug(
                        f"Used fallback reservoir sampling for {table}.{col} ({len(samples)} samples)"
                    )

                # assemble field info
                field_info: Dict[str, Any] = {
                    "data_type": typ,
                    "data": samples
                }
                if typ.lower() in DATETIME_TYPES:
                    field_info["datetime_format"] = infer_date_format(samples)

                schema[table][col] = field_info

        # persist schema
        os.makedirs(save_dir, exist_ok=True)
        out_path = os.path.join(save_dir, f"{schema_name}.pkl")
        with open(out_path, "wb") as f:
            pickle.dump(dict(schema), f)
        logger.info(f"Schema pickle written to {out_path}")

        if save_json_version:
            json_dir = os.path.dirname(save_json_version)
            if json_dir:
                os.makedirs(json_dir, exist_ok=True)
            schema_dict = json.loads(json.dumps(dict(schema), default=str))
            with open(save_json_version, "w", encoding="utf-8") as jf:
                json.dump(schema_dict, jf, indent=2)
            logger.info(f"Schema JSON written to {save_json_version}")

        return out_path


    def list_databases(
        self,
        show_fields: bool = False,
        show_system_data: bool = False
    ) -> dict:
        conn = self.connection_manager.connect()
        schema: Dict[str, Any] = {}

        with get_cursor(conn) as cursor:
            cursor.execute("SHOW DATABASES")
            dbs = [row[0] for row in cursor.fetchall()]

            if not show_system_data:
                system_dbs = {"information_schema", "mysql", "performance_schema", "sys"}
                dbs = [db for db in dbs if db not in system_dbs]

            for db in dbs:
                cursor.execute(f"SHOW TABLES FROM `{db}`")
                tables = [row[0] for row in cursor.fetchall()]

                if show_fields:
                    table_info: Dict[str, List[str]] = {}
                    for tbl in tables:
                        cursor.execute(f"SHOW COLUMNS FROM `{db}`.`{tbl}`")
                        cols = [col[0] for col in cursor.fetchall()]
                        table_info[tbl] = cols
                    schema[db] = table_info
                else:
                    schema[db] = tables

        logger.info(
            f"Retrieved schema for {len(schema)} databases "
            f"(show_fields={show_fields}, show_system_data={show_system_data})"
        )
        return schema

    def execute_query(
        self, 
        sql_query: str, 
        limit: Optional[int] = 100
    ) -> Dict[str, Any]:
        """
        Executes a given SQL query and returns the result.
        If a limit is provided, it will be applied intelligently. If the original query
        has a stricter (smaller) limit, it will be respected.
        
        Args:
            sql_query: The SQL query to execute
            limit: Optional limit to apply to the query
            
        Returns:
            Dictionary containing the result data, error (if any), and final SQL
        """
        conn = self.connection_manager.connect()
        params: List[Any] = []

        if limit is not None:
            limit_pattern = re.compile(r'LIMIT\s+(\d+)\s*;?\s*$', re.IGNORECASE)
            match = limit_pattern.search(sql_query)

            if match:
                existing_limit = int(match.group(1))
                if limit < existing_limit:
                    sql_query = limit_pattern.sub('LIMIT %s', sql_query)
                    params.append(limit)
                    logger.info(
                        f"Replacing existing LIMIT {existing_limit} with new, stricter LIMIT {limit}"
                    )
                else:
                    logger.info(
                        f"Respecting existing LIMIT {existing_limit} as it is stricter than requested LIMIT {limit}"
                    )
            else:
                sql_query = sql_query.rstrip().rstrip(";") + " LIMIT %s"
                params.append(limit)
                logger.info(f"Appending new LIMIT {limit} to query")

        try:
            with get_cursor(conn, dictionary=True) as cursor:
                cursor.execute(sql_query, tuple(params))
                result = cursor.fetchall()

            # Build a display version of the SQL
            final_sql = sql_query
            if params:
                temp_sql = sql_query
                for param in params:
                    temp_sql = temp_sql.replace('%s', str(param), 1)
                final_sql = temp_sql

            return {"data": result, "error": None, "final_sql": final_sql}
        except mysql.connector.Error as e:
            logger.error(f"Error executing query: {e}")
            return {"data": None, "error": str(e), "final_sql": sql_query}

    def ensure_transform_functions(
        self,
        initialism_fn: str = "initialism",
        metaphone_udf: str = "double_metaphone",
        udf_soname: str = "lib_mysqludf_morph.so"
    ) -> None:
        """
        Ensure that the MySQL functions for initialism (stored function) and metaphone (UDF)
        exist in the connected database. Creates them if missing.
        """
        conn = self.connection_manager.connect()
        try:
            with get_cursor(conn) as cursor:
                # Check for existing stored function
                cursor.execute(
                    "SELECT COUNT(*) FROM information_schema.routines "
                    "WHERE routine_schema=%s AND routine_name=%s AND routine_type='FUNCTION'",
                    (self.connection_manager.config.database, initialism_fn)
                )
                if cursor.fetchone()[0] == 0:
                    # Drop any existing stub
                    cursor.execute(f"DROP FUNCTION IF EXISTS `{initialism_fn}`")
                    conn.commit()
                    # Use a pure-Python connection to execute multi-statement DDL
                    cfg = self.connection_manager.config
                    pure_conn = mysql.connector.connect(
                        host=cfg.host,
                        user=cfg.user,
                        password=cfg.password,
                        port=cfg.port,
                        database=cfg.database,
                        use_pure=True
                    )
                    try:
                        pure_cursor = pure_conn.cursor()
                        create_fn_sql = f"""
                        CREATE FUNCTION `{initialism_fn}`(txt TEXT)
                          RETURNS VARCHAR(255)
                          DETERMINISTIC
                        BEGIN
                          DECLARE result VARCHAR(255) DEFAULT '';
                          DECLARE word  VARCHAR(255);
                          DECLARE pos   INT DEFAULT 1;
                          WHILE pos > 0 DO
                            SET word = SUBSTRING_INDEX(SUBSTRING(txt, pos), ' ', 1);
                            IF word = '' THEN
                              LEAVE;
                            END IF;
                            SET result = CONCAT(result, UPPER(LEFT(word, 1)));
                            SET pos = LOCATE(' ', txt, pos) + 1;
                            IF pos = 1 THEN
                              LEAVE;
                            END IF;
                          END WHILE;
                          RETURN result;
                        END;
                        """
                        pure_cursor.execute(create_fn_sql, multi=True)
                        pure_conn.commit()
                        logger.info(f"Created stored function `{initialism_fn}`")
                    finally:
                        pure_conn.close()
                # Check for existing UDF
                cursor.execute(
                    "SELECT COUNT(*) FROM mysql.func WHERE name = %s",
                    (metaphone_udf,)
                )
                if cursor.fetchone()[0] == 0:
                    cursor.execute(f"INSTALL SONAME '{udf_soname}'")
                    conn.commit()
                    logger.info(f"Installed UDF library for `{metaphone_udf}`")
        finally:
            pass


from typing import List, Optional
from enum import Enum
from pydantic import BaseModel

class CreateDDLResponse(BaseModel):
    ddl: str
    ddl_path: Optional[str] = None

class FieldTransformType(str, Enum):
    INITIALISM = "initialism"
    PHONETIC = "phonetic"
    SOUNDEX = "soundex"

class FieldTransform(BaseModel):
    original_field_name: str
    transformed_field_name: str
    field_transform: FieldTransformType

class TableField(BaseModel):
    table_name: str
    field_name: str

class EntityMappingRequest(BaseModel):
    entity_name: str
    select_field: TableField
    top_n: int = 20
    max_prefilter: int = 1000

class SingleMappedEntity(BaseModel):
    raw_entity: str
    canonical_entity: str
    canonical_table_name: str
    canonical_field_name: str
    score: float

class EntityMappingResponse(BaseModel):
    results: List[SingleMappedEntity]



# canonmap/services/db_mysql/helpers/create_ddl.py

import re
from decimal import Decimal
from typing import List, Optional
from pydantic import BaseModel

import pandas as pd


class CreateDDLResponse(BaseModel):
    ddl: str
    ddl_path: Optional[str] = None

def create_mysql_ddl(csv_path_or_buffer, table_name, save_dir=None):
    """
    Generate a MySQL CREATE TABLE DDL statement from a CSV file.
    
    Parameters:
        csv_path_or_buffer: Path to the CSV file (or a file-like object).
        table_name (str): Desired table name (will be cleaned for MySQL).
        save_dir (str or Path, optional): File path to save the DDL statement.
    
    Returns:
        str: The CREATE TABLE statement corresponding to the CSV’s schema.
    """
    # Load CSV with all columns as strings to preserve data for analysis
    df = pd.read_csv(
        csv_path_or_buffer,
        dtype=str,
        keep_default_na=False,
        na_values=[],
        engine='python',
        on_bad_lines='skip'
    )
    
    # Clean table name to be a valid MySQL identifier
    table = table_name.strip()
    table = re.sub(r'[^0-9A-Za-z_]', '_', table)            # replace invalid chars with _
    if table == '' or table.isdigit() or re.match(r'^\d', table):
        table = 'tbl_' + table                              # prefix if starts with digit or is empty
    table = table.lower()
    # Avoid reserved word conflict
    mysql_reserved = {
        "add","all","alter","analyze","and","as","asc","between","by","create","delete","desc","distinct",
        "drop","exists","from","group","having","in","index","insert","into","join","key","like","limit",
        "not","null","on","or","order","outer","select","set","table","update","union","where"
    }
    if table in mysql_reserved:
        table += "_tbl"
    if len(table) > 64:
        table = table[:64]
    
    # Clean column names similarly
    col_name_map = {}
    used_names = set()
    for col in df.columns:
        col_clean = col.strip()
        col_clean = re.sub(r'[^0-9A-Za-z_]', '_', col_clean)
        if col_clean == '' or col_clean.isdigit() or re.match(r'^\d', col_clean):
            col_clean = 'col_' + col_clean
        col_clean = col_clean.lower()
        if col_clean in mysql_reserved:
            col_clean += "_col"
        if len(col_clean) > 64:
            col_clean = col_clean[:64]
        # Ensure uniqueness by appending a number if needed
        base_name = col_clean
        count = 1
        while col_clean in used_names:
            col_clean = f"{base_name}_{count}"
            count += 1
            if len(col_clean) > 64:
                # truncate base part if necessary to keep length ≤ 64
                trim_len = len(col_clean) - 64
                base_name = base_name[:-trim_len]
                col_clean = f"{base_name}_{count}"
        used_names.add(col_clean)
        col_name_map[col] = col_clean
    
    # Prepare a list for column definitions
    column_defs = []
    # Define placeholders that indicate missing data
    missing_values = {"", None, "NULL", "null", "NA", "na", "N/A", "n/a"}
    
    for col in df.columns:
        clean_name = col_name_map[col]
        values = df[col].tolist()
        # Filter out missing placeholders for type inference
        non_null_vals = [v for v in values if v not in missing_values]
        if not non_null_vals:
            # If column is entirely empty, default to TEXT (could also be VARCHAR(1) as minimal)
            col_type = "TEXT"
        else:
            # Check for booleans (true/false, yes/no, 0/1)
            val_set = {str(v).strip().lower() for v in non_null_vals}
            boolean_vals = {"true", "false", "t", "f", "yes", "no", "y", "n", "0", "1"}
            if val_set <= boolean_vals:
                col_type = "TINYINT(1)"  # MySQL boolean
            else:
                # Check if all values are numeric (int or float) or can be numeric
                all_numeric = True
                is_integer = True
                min_val = None
                max_val = None
                max_int_digits = 0
                max_frac_digits = 0
                for val in non_null_vals:
                    s = str(val).strip()
                    # General number (int/float) regex
                    if not re.fullmatch(r'-?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?', s):
                        all_numeric = False
                        break
                    # If matches, determine if it has a fractional part or exponent
                    if re.fullmatch(r'-?\d+', s):
                        # pure integer string
                        n = int(s)
                        # track min/max for range
                        if min_val is None or n < min_val: 
                            min_val = n
                        if max_val is None or n > max_val: 
                            max_val = n
                        # count significant digits of integer
                        int_len = len(s.lstrip('-').lstrip('0')) or 1
                        if int_len > max_int_digits:
                            max_int_digits = int_len
                    else:
                        # number with decimal point or exponent
                        is_integer = False
                        try:
                            dec = Decimal(s)
                        except Exception:
                            dec = Decimal(str(float(s)))  # fallback to float if Decimal fails
                        # Convert to normalized string without scientific notation
                        dec_str = format(dec, 'f')
                        if dec_str.startswith('-'):
                            dec_str = dec_str[1:]
                        if '.' in dec_str:
                            int_part, frac_part = dec_str.split('.')
                        else:
                            int_part, frac_part = dec_str, ''
                        # Count digits in int and frac parts (ignore trailing zeros in fraction)
                        int_part = int_part.lstrip('0')
                        int_len = len(int_part) if int_part != '' else 0
                        frac_len = len(frac_part.rstrip('0'))
                        if int_len == 0 and frac_len == 0:
                            # value is 0.0 or 0.000... 
                            int_len = 1
                        if int_len > max_int_digits:
                            max_int_digits = int_len
                        if frac_len > max_frac_digits:
                            max_frac_digits = frac_len
                        # track min/max as floats (for range if needed)
                        try:
                            n = float(dec)
                        except Exception:
                            n = None
                        if n is not None:
                            if min_val is None or n < min_val:
                                min_val = n
                            if max_val is None or n > max_val:
                                max_val = n
                if all_numeric:
                    # If numeric, decide between int vs decimal
                    if is_integer:
                        # Choose appropriate integer type based on range
                        signed = True
                        if min_val is not None and min_val >= 0:
                            signed = False
                        if signed:
                            # use smallest signed type that fits min_val...max_val
                            if min_val is not None and max_val is not None:
                                if min_val >= -128 and max_val <= 127:
                                    col_type = "TINYINT"
                                elif min_val >= -32768 and max_val <= 32767:
                                    col_type = "SMALLINT"
                                elif min_val >= -8388608 and max_val <= 8388607:
                                    col_type = "MEDIUMINT"
                                elif min_val >= -2147483648 and max_val <= 2147483647:
                                    col_type = "INT"
                                elif min_val >= -9223372036854775808 and max_val <= 9223372036854775807:
                                    col_type = "BIGINT"
                                else:
                                    # If out of BIGINT range, use DECIMAL with all digits
                                    col_type = f"DECIMAL({max_int_digits},0)"
                            else:
                                col_type = "INT"
                        else:
                            # all values >= 0, use unsigned types
                            if max_val is not None:
                                if max_val <= 255:
                                    col_type = "TINYINT UNSIGNED"
                                elif max_val <= 65535:
                                    col_type = "SMALLINT UNSIGNED"
                                elif max_val <= 16777215:
                                    col_type = "MEDIUMINT UNSIGNED"
                                elif max_val <= 4294967295:
                                    col_type = "INT UNSIGNED"
                                elif max_val <= 18446744073709551615:
                                    col_type = "BIGINT UNSIGNED"
                                else:
                                    col_type = f"DECIMAL({max_int_digits},0)"
                            else:
                                col_type = "INT UNSIGNED"
                    else:
                        # Numeric column with a decimal part or very large/small values
                        if max_frac_digits == 0:
                            # (This case: all_numeric true, but is_integer false could happen if scientific notation of an int)
                            # Treat as integer range in such rare case
                            col_type = "INT"  # (fallback, though we could refine similar to above int logic)
                        else:
                            # Use DECIMAL for exact representation of floats
                            precision = max_int_digits + max_frac_digits
                            if precision > 65:
                                # Too many digits for DECIMAL, use floating type
                                col_type = "DOUBLE"
                            else:
                                scale = max_frac_digits
                                col_type = f"DECIMAL({precision},{scale})"
                else:
                    # Not all values are numeric – treat as text
                    max_len = max(len(str(v)) for v in non_null_vals)
                    if max_len <= 255:
                        col_type = f"VARCHAR({max_len})"
                    elif max_len <= 65535:
                        col_type = "TEXT"
                    elif max_len <= 16777215:
                        col_type = "MEDIUMTEXT"
                    else:
                        col_type = "LONGTEXT"
        # Append the column definition
        column_defs.append(f"  `{clean_name}` {col_type}")
    
    # Construct the CREATE TABLE statement
    ddl = "CREATE TABLE `{}` (\n".format(table)
    ddl += ",\n".join(column_defs)
    ddl += "\n);"
    
    # Save to file if requested
    if save_dir:
        with open(save_dir, 'w') as f:
            f.write(ddl)

    return CreateDDLResponse(ddl=ddl, ddl_path=save_dir)


import logging
import csv
from pathlib import Path
from typing import Union, List, Dict, Any, Tuple
import pandas as pd

from canonmap.services.db_mysql.adapters.connection import ConnectionManager
from canonmap.services.db_mysql.helpers.create_ddl import create_mysql_ddl

logger = logging.getLogger(__name__)

class CSVManager:
    def __init__(self, connection_manager: ConnectionManager):
        self.connection_manager = connection_manager

    def create_table_ddl_from_csv(
        self,
        csv_path_or_buffer,
        table_name: str,
        save_dir: Union[None, str, Path] = None
    ) -> str:
        return create_mysql_ddl(csv_path_or_buffer, table_name, save_dir)

    def _smart_fix_csv(
        self,
        csv_path: Union[str, Path],
        fixed_csv_path: Union[str, Path] = "fixed.csv",
        bad_rows_path: Union[str, Path] = "bad_rows.csv",
        log_bad_rows: bool = True
    ) -> Tuple[str, List[Tuple[int, str]]]:
        """
        Reads a CSV, identifies and separates malformed rows, and writes a clean version.
        A row is considered malformed if it has a different number of columns than the header.
        It attempts to fix rows with more columns by merging adjacent fields.
        
        Returns:
            Tuple[str, List[Tuple[int, str]]]: (fixed_csv_path, bad_rows_data)
        """
        csv_path = Path(csv_path)
        fixed_csv_path = Path(fixed_csv_path)
        bad_rows_path = Path(bad_rows_path)
        
        bad_rows_data = []

        try:
            with open(csv_path, 'r', newline='', encoding='utf-8') as infile:
                # Read header to determine expected number of columns
                try:
                    header = next(csv.reader(infile))
                    expected_cols = len(header)
                except StopIteration:
                    logger.warning(f"CSV file '{csv_path}' is empty.")
                    # Create empty fixed file and return
                    fixed_csv_path.touch()
                    return str(fixed_csv_path), []

                # Rewind and process whole file
                infile.seek(0)
                reader = csv.reader(infile)
                next(reader) # Skip header

                with open(fixed_csv_path, 'w', newline='', encoding='utf-8') as outfile:
                    writer = csv.writer(outfile)
                    writer.writerow(header)

                    for lineno, row in enumerate(reader, start=2):
                        if len(row) == expected_cols:
                            writer.writerow(row)
                        elif len(row) > expected_cols:
                            # Heuristic: try to merge fields to fix it.
                            # This merges the second field with subsequent fields to meet the expected column count.
                            num_extra_cols = len(row) - expected_cols
                            merged_row = row[:1] + [','.join(row[1:1 + num_extra_cols + 1])] + row[1 + num_extra_cols + 1:]
                            if len(merged_row) == expected_cols:
                                writer.writerow(merged_row)
                                logger.debug(f"Fixed line {lineno} by merging columns.")
                            else:
                                # This case should ideally not be reached with this logic
                                bad_rows_data.append((lineno, ','.join(row)))
                        else: # len(row) < expected_cols
                            # This is a malformed row, log it. 
                            # Don't pad, as it can hide data integrity issues.
                            bad_rows_data.append((lineno, ','.join(row)))
            
            if bad_rows_data and log_bad_rows:
                logger.warning(f"Found {len(bad_rows_data)} malformed row(s). Writing them to '{bad_rows_path}'.")
                with open(bad_rows_path, 'w', newline='', encoding='utf-8') as bad_file:
                    writer = csv.writer(bad_file)
                    writer.writerow(['line_number', 'original_content'])
                    for lineno, content in bad_rows_data:
                        writer.writerow([lineno, content])

            logger.info(f"✅ Finished cleaning. Clean CSV written to: {fixed_csv_path}")
            return str(fixed_csv_path), bad_rows_data

        except Exception as e:
            logger.error(f"Error processing CSV file '{csv_path}': {e}")
            raise

    def process_csv_to_json(
        self,
        csv_path: Union[str, Path],
        table_name: str
    ) -> Tuple[List[Dict[str, Any]], str]:
        """
        Process a CSV file: clean it, convert to JSON format, and generate DDL statement.
        
        Args:
            csv_path (Union[str, Path]): Path to the CSV file
            table_name (str): Name of the table for DDL generation
            
        Returns:
            Tuple[List[Dict[str, Any]], str]: (json_data, ddl_statement)
        """
        try:
            # Use a more descriptive name for the bad rows file.
            bad_rows_csv_path = Path(csv_path).stem + "_bad_rows.csv"
            fixed_csv_path, bad_rows = self._smart_fix_csv(
                csv_path, 
                bad_rows_path=bad_rows_csv_path,
                log_bad_rows=True
            )
            
            if not Path(fixed_csv_path).exists() or Path(fixed_csv_path).stat().st_size == 0:
                logger.warning(f"CSV file '{csv_path}' is empty or could not be processed. Nothing to load.")
                return [], ""

            # Generate DDL statement
            ddl_statement = self.create_table_ddl_from_csv(fixed_csv_path, table_name)
            
            # Convert CSV to JSON format
            import pandas.errors
            logger.info(f"Converting CSV '{fixed_csv_path}' to JSON format...")
            try:
                # Read the CSV, keeping blank values as empty strings and handling NaNs
                df = pd.read_csv(fixed_csv_path, keep_default_na=False, na_values=[''])

                # Replace any numpy NaN values with None for SQL compatibility (NULL)
                df = df.astype(object).where(pd.notnull(df), None)
                
                # Convert DataFrame to list of dictionaries
                json_data = df.to_dict('records')
                
                logger.info(f"✅ Converted {len(json_data)} rows to JSON format")
                return json_data, ddl_statement

            except pandas.errors.ParserError as e:
                logger.error(f"ParserError while reading cleaned CSV '{fixed_csv_path}': {e}")
                raise
            except Exception as e:
                logger.error(f"An unexpected error occurred while reading the CSV: {e}")
                raise
            
        except Exception as e:
            logger.error(f"Error while processing CSV '{csv_path}': {e}")
            raise


# canonmap/services/db_mysql/adapters/connection.py

import logging
from typing import Optional

import mysql.connector
from mysql.connector import MySQLConnection, Error

from canonmap.services.db_mysql.config import CanonMapMySQLConfig

logger = logging.getLogger(__name__)

class ConnectionManager:
    def __init__(self, config: CanonMapMySQLConfig):
        self.config = config
        self.conn: Optional[MySQLConnection] = None

    def connect(self) -> MySQLConnection:
        if self.conn and self.conn.is_connected():
            logger.debug("Reusing existing MySQL connection")
            return self.conn

        config = {
            "user": self.config.user,
            "password": self.config.password,
        }

        # Use Unix socket if provided (for Cloud SQL)
        if self.config.unix_socket:
            config["unix_socket"] = self.config.unix_socket
            logger.info(f"Using Unix socket connection: {self.config.unix_socket}")
        else:
            # Use TCP connection
            config.update({
                "host": self.config.host,
                "port": self.config.port,
            })
            logger.info(f"Using TCP connection to {self.config.host}:{self.config.port}")

        try:
            self.conn = mysql.connector.connect(**config, allow_local_infile=True)
            if self.config.database:
                self.conn.database = self.config.database
            logger.info(f"Connected to MySQL server successfully")
            return self.conn

        except Error as e:
            logger.error("Could not connect to MySQL: %s", e, exc_info=True)
            raise RuntimeError(f"Could not connect to MySQL: {e}") from e

    def close(self) -> None:
        if self.conn and self.conn.is_connected():
            self.conn.close()
            self.conn = None
            logger.info("MySQL connection closed")

    def connection_info(self) -> dict:
        status = "Connected" if (self.conn and self.conn.is_connected()) else "Disconnected"
        info = {
            "user": self.config.user,
            "database": self.config.database,
            "status": status,
        }
        
        if self.config.unix_socket:
            info["unix_socket"] = self.config.unix_socket
        else:
            info.update({
                "host": self.config.host,
                "port": self.config.port,
            })
            
        # log a prettified multi-line summary
        for k, v in info.items():
            logger.info(f"  {k}: {v}")
        return info


import logging

from canonmap.services.db_mysql.schemas import EntityMappingRequest

logger = logging.getLogger(__name__)

try:
    import Levenshtein  # pip install python-Levenshtein
    _have_lev = True
except ImportError:
    from difflib import SequenceMatcher
    _have_lev = False


def search_levenshtein(conn, request: EntityMappingRequest):
    entity_name = request.entity_name
    threshold = request.threshold
    max_prefilter = request.max_prefilter

    for field_object in request.select_fields:
        field_name = field_object.field_name
        table_name = field_object.table_name
        
        # Create a more flexible prefix by taking the first word and making it case-insensitive
        entity_name_split = entity_name.split(" ")
        if len(entity_name_split[0]) >= 2:
            # Use first 2+ characters of first word for broader matching
            prefix = entity_name_split[0][:2].lower() + '%'
        else:
            prefix = entity_name_split[0].lower() + '%'
        
        # Use case-insensitive LIKE for prefix matching and select the actual field values
        sql = """
        SELECT `{field_name}`
        FROM `{table_name}`
        WHERE LOWER(`{field_name}`) LIKE %s
        LIMIT %s;
        """
        sql = sql.format(table_name=table_name, field_name=field_name)
        cur = conn.cursor(buffered=True)
        cur.execute(sql, (prefix, max_prefilter))
        candidates = cur.fetchall()
        logger.info(f"Found {len(candidates)} candidates for prefix '{prefix}'")
        
        # normalize query once
        name_norm = entity_name.lower().replace(" ", "")

        results = []
        for row in candidates:
            cand = row[0]  # Get the field value from the first column
            # normalize candidate
            cand_norm = cand.lower().replace(" ", "")
            if _have_lev:
                score = Levenshtein.ratio(name_norm, cand_norm)
            else:
                score = SequenceMatcher(None, name_norm, cand_norm).ratio()
            if score >= threshold:
                results.append((cand, round(score, 3)))
        
        # If we have candidates but no results and threshold > 0, try with threshold = 0.0
        if len(candidates) > 0 and len(results) == 0 and threshold > 0.0:
            logger.info(f"No results found with threshold {threshold}, retrying with threshold 0.0")
            request_copy = EntityMappingRequest(
                entity_name=request.entity_name,
                select_fields=request.select_fields,
                threshold=0.0,
                max_prefilter=request.max_prefilter
            )
            return search_levenshtein(conn, request_copy)
        
        # sort by best match first
        results.sort(key=lambda x: x[1], reverse=True)
        return results




# canonmap/services/db_mysql/adapters/cursor.py

import logging
from contextlib import contextmanager
from mysql.connector import MySQLConnection, Error

logger = logging.getLogger(__name__)

@contextmanager
def get_cursor(
    connection: MySQLConnection,
    dictionary: bool = False
):
    """
    Context manager for MySQL cursors.
    Yields a cursor and ensures it’s closed afterwards,
    logging any execution errors.
    
    :param connection: an open MySQLConnection
    :param dictionary: if True, cursor returns dicts instead of tuples
    """
    cursor = connection.cursor(dictionary=dictionary)
    try:
        yield cursor
    except Error as e:
        logger.error("Cursor operation failed: %s", e, exc_info=True)
        # Roll back any uncommitted transaction if desired:
        try:
            connection.rollback()
        except Error:
            pass
        raise
    finally:
        cursor.close()


# canonmap/core.py
import logging

from canonmap.services.db_mysql.core import CanonMapMySQLClient
from canonmap.utils.logger import configure_logging

configure_logging("INFO")
logger = logging.getLogger(__name__)

class CanonMap:
    def __init__(self, mysql_client: CanonMapMySQLClient):
        self.mysql_client = mysql_client



# canonmap/services/db_mysql/managers/table.py

import logging
import re
from pathlib import Path
from typing import Union, Optional, List, Dict, Any

from canonmap.services.db_mysql.adapters.connection import ConnectionManager
from canonmap.services.db_mysql.schemas import TableField, FieldTransformType
from canonmap.services.db_mysql.adapters.cursor import get_cursor

from canonmap.services.db_mysql.schemas import CreateDDLResponse # type: ignore
from canonmap.services.db_mysql.managers.database import DatabaseManager

logger = logging.getLogger(__name__)

class TableManager:
    def __init__(self, connection_manager: ConnectionManager):
        self.connection_manager = connection_manager

    #########################################################
    # Table Creation
    #########################################################
    def create_table_from_ddl(
        self,
        ddl: Union[str, Path, CreateDDLResponse],
        if_exists: str = 'skip',
    ) -> None:
        ddl_sql = ddl
        try:
            path = Path(ddl)
            if path.exists():
                if path.suffix.lower() != '.sql':
                    raise ValueError(
                        f"Invalid file extension for DDL file: {ddl!r}. Must end with '.sql'."
                    )
                ddl_sql = path.read_text()
        except (OSError, TypeError):
            # Check if it's a CreateDDLResponse object
            if hasattr(ddl, 'ddl'):
                ddl_sql = ddl.ddl
            else:
                ddl_sql = ddl  # keep original if not a file

        conn = self.connection_manager.connect()
        statements = [stmt.strip() for stmt in ddl_sql.split(';') if stmt.strip()]
        
        tables_created = []
        tables_skipped = []
        
        with get_cursor(conn) as cursor:
            for stmt in statements:
                # Check if this is a CREATE TABLE statement
                if stmt.strip().upper().startswith('CREATE TABLE'):
                    # Extract table name from CREATE TABLE statement
                    # table_name = self._extract_table_name(stmt)
                    match = re.search(r'CREATE\s+TABLE\s+(?:IF\s+NOT\s+EXISTS\s+)?[`"]?(\w+)[`"]?\s*\(', stmt, re.IGNORECASE)
                    table_name = match.group(1) if match else None
                    if table_name:
                        # Check if table exists
                        cursor.execute(
                            "SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES "
                            "WHERE TABLE_SCHEMA = %s AND TABLE_NAME = %s",
                            (self.connection_manager.config.database, table_name)
                        )
                        table_exists = cursor.fetchone() is not None
                        
                        if table_exists:
                            if if_exists == 'skip':
                                logger.info(f"Table '{table_name}' already exists, skipping creation.")
                                tables_skipped.append(table_name)
                                continue
                            elif if_exists == 'replace':
                                logger.info(f"Table '{table_name}' already exists, dropping and recreating.")
                                cursor.execute(f"DROP TABLE `{table_name}`")
                                tables_created.append(table_name)
                            elif if_exists == 'error':
                                raise ValueError(f"Table '{table_name}' already exists. Use if_exists='skip' or 'replace' to handle this.")
                        else:
                            tables_created.append(table_name)
                
                # Execute the statement
                cursor.execute(stmt)
        
        conn.commit()
        
        # Log clear summary of what happened
        if tables_created:
            logger.info(f"Successfully created tables: {', '.join(tables_created)}")
        if tables_skipped:
            logger.info(f"Skipped existing tables: {', '.join(tables_skipped)}")
        if not tables_created and not tables_skipped:
            logger.info("No table creation statements found in DDL.")
    
    def create_table_from_data(
        self,
        data: List[Dict[str, Any]],
        table_name: str,
        if_exists: str = "append",
        chunk_size: int = 1000,
        ddl_statement: Optional[str] = None
    ):
        if not data:
            logger.warning(f"No data provided to load into table '{table_name}'. Nothing to load.")
            return

        conn = self.connection_manager.connect()
        with get_cursor(conn) as cursor:
            # Check if table exists
            cursor.execute(
                "SELECT 1 FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA=%s AND TABLE_NAME=%s",
                (self.connection_manager.config.database, table_name),
            )
            table_exists = cursor.fetchone() is not None

            # Handle if_exists logic
            if table_exists:
                if if_exists == "replace":
                    logger.info(f"Table '{table_name}' exists and if_exists is 'replace'. Dropping and recreating table.")
                    cursor.execute(f"DROP TABLE IF EXISTS `{table_name}`")
                    conn.commit()
                    if ddl_statement:
                        self.create_table_from_ddl(ddl_statement)
                        logger.info(f"Table '{table_name}' recreated.")
                    else:
                        raise ValueError(f"DDL statement required when if_exists='replace' and table doesn't exist.")
                elif if_exists == "fail":
                    raise ValueError(f"Table '{table_name}' already exists. Set if_exists to 'append' or 'replace'.")
                # if 'append', do nothing to the table structure
            
            elif not table_exists:
                if if_exists in ["append", "fail"]:
                    logger.info(f"Table '{table_name}' does not exist. Creating it now.")
                    if ddl_statement:
                        self.create_table_from_ddl(ddl_statement)
                        logger.info(f"Table '{table_name}' created.")
                    else:
                        raise ValueError(f"DDL statement required when table doesn't exist.")

            # Get column names from first row
            columns = list(data[0].keys())
            placeholders = ", ".join(["%s"] * len(columns))
            columns_sql = ", ".join([f"`{col}`" for col in columns])

            # Insert data in chunks
            total_inserted = 0
            for start in range(0, len(data), chunk_size):
                chunk = data[start : start + chunk_size]
                # Convert each row to tuple in the same order as columns
                chunk_data = [tuple(row[col] for col in columns) for row in chunk]
                sql = f"INSERT INTO `{table_name}` ({columns_sql}) VALUES ({placeholders})"
                cursor.executemany(sql, chunk_data)
                total_inserted += len(chunk_data)
            
            conn.commit()
            logger.info(f"Inserted {total_inserted} rows into '{table_name}'.")


    #########################################################
    # Table TableField Management
    #########################################################
    def create_table_fields(
        self,
        fields: list["TableField"],
        field_transform: FieldTransformType,
        pk_field: str | None = None,
        batch_size: int = 10_000,
        skip_existing: bool = True,
    ) -> dict[str, list[str]]:
        """
        For each TableField create a derived column with the specified transform:
        initialism  -> FIRST LETTERS ONLY (ABC)
        phonetic    -> Double Metaphone primary code
        soundex     -> MySQL SOUNDEX()

        Args:
            fields: List of TableField objects to transform
            field_transform: Type of transform to apply to all fields
            pk_field: Primary key field name (auto-detected if None)
            batch_size: Batch size for processing large tables
            skip_existing: If True, only process NULL/empty cells in transformed fields

        Returns: {table_name: [new_field_names]}
        """
        # ensure the transform functions exist for initialism and phonetic
        if field_transform in (FieldTransformType.INITIALISM, FieldTransformType.PHONETIC):
            db_manager = DatabaseManager(self.connection_manager)
            db_manager.ensure_transform_functions()
        # Validate transform type
        if field_transform not in {FieldTransformType.INITIALISM, FieldTransformType.PHONETIC, FieldTransformType.SOUNDEX}:
            raise ValueError(f"Invalid transform '{field_transform}'")
        
        # Group by table
        by_table: dict[str, list[TableField]] = {}
        for f in fields:
            by_table.setdefault(f.table_name, []).append(f)

        conn = self.connection_manager.connect()

        import math, re
        try:
            from metaphone import doublemetaphone
        except ImportError:
            doublemetaphone = None

        def to_initialism(text: str | None) -> str | None:
            if not text:
                return None
            parts = re.findall(r"[A-Za-z]+", text)
            return "".join(p[0].upper() for p in parts) if parts else None

        def to_phonetic(text: str | None) -> str | None:
            if not text:
                return None
            if doublemetaphone is None:
                raise RuntimeError("metaphone package not installed")
            p, s = doublemetaphone(text)
            return p or s or None

        created: dict[str, list[str]] = {}

        for table_name, flist in by_table.items():
            # Decide PK (use provided or auto)
            effective_pk = pk_field or self._get_or_create_table_pk(table_name)

            # Fetch columns
            with get_cursor(conn) as cursor:
                cursor.execute(
                    "SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS "
                    "WHERE TABLE_SCHEMA=%s AND TABLE_NAME=%s",
                    (self.connection_manager.config.database, table_name)
                )
                cols = {r[0] for r in cursor.fetchall()}

            # Split work
            sql_jobs: list[tuple[TableField, str]] = []
            py_jobs:  list[tuple[TableField, str]] = []

            for f in flist:
                if f.field_name not in cols:
                    raise ValueError(f"Column '{f.field_name}' not found in '{table_name}'")

                new_field = f"__{f.field_name}_{field_transform.value.lower()}__"
                if new_field not in cols:
                    with get_cursor(conn) as cursor:
                        cursor.execute(f"ALTER TABLE `{table_name}` ADD COLUMN `{new_field}` VARCHAR(255)")
                    conn.commit()
                    cols.add(new_field)

                if field_transform == FieldTransformType.SOUNDEX:
                    sql_jobs.append((f, new_field))
                else:
                    py_jobs.append((f, new_field))

            # SOUNDEX via SQL
            for f, new_field in sql_jobs:
                with get_cursor(conn) as cursor:
                    if skip_existing:
                        cursor.execute(
                            f"UPDATE `{table_name}` SET `{new_field}` = SOUNDEX(`{f.field_name}`) "
                            f"WHERE `{new_field}` IS NULL OR `{new_field}` = ''"
                        )
                    else:
                        cursor.execute(
                            f"UPDATE `{table_name}` SET `{new_field}` = SOUNDEX(`{f.field_name}`)"
                        )
                conn.commit()

            # Python transforms batched
            if py_jobs:
                # Count total rows to process
                if skip_existing:
                    # Count only rows where transformed fields are NULL/empty
                    null_conditions = []
                    for f, new_field in py_jobs:
                        null_conditions.append(f"`{new_field}` IS NULL OR `{new_field}` = ''")
                    
                    with get_cursor(conn) as cursor:
                        cursor.execute(f"SELECT COUNT(*) FROM `{table_name}` WHERE {' OR '.join(null_conditions)}")
                        total_rows = cursor.fetchone()[0]
                else:
                    with get_cursor(conn) as cursor:
                        cursor.execute(f"SELECT COUNT(*) FROM `{table_name}`")
                        total_rows = cursor.fetchone()[0]

                batches = math.ceil(total_rows / batch_size)
                select_cols = ", ".join({f"`{effective_pk}`"} | {f"`{f.field_name}`" for f, _ in py_jobs})

                for i in range(batches):
                    offset = i * batch_size
                    with get_cursor(conn) as cursor:
                        if skip_existing:
                            # Only select rows where transformed fields are NULL/empty
                            null_conditions = []
                            for f, new_field in py_jobs:
                                null_conditions.append(f"`{new_field}` IS NULL OR `{new_field}` = ''")
                            
                            cursor.execute(
                                f"SELECT {select_cols} FROM `{table_name}` "
                                f"WHERE {' OR '.join(null_conditions)} "
                                f"LIMIT %s OFFSET %s",
                                (batch_size, offset)
                            )
                        else:
                            cursor.execute(
                                f"SELECT {select_cols} FROM `{table_name}` LIMIT %s OFFSET %s",
                                (batch_size, offset)
                            )
                        rows = cursor.fetchall()
                        # capture column order once
                        colnames = [d[0] for d in cursor.description]

                    if not rows:
                        break

                    updates: dict[str, list[tuple[str | None, object]]] = {nf: [] for _, nf in py_jobs}
                    pk_idx = colnames.index(effective_pk)
                    for row in rows:
                        pk_val = row[pk_idx]
                        row_map = dict(zip(colnames, row))
                        for f, new_field in py_jobs:
                            src_val = row_map[f.field_name]
                            if field_transform == FieldTransformType.INITIALISM:
                                transformed = to_initialism(src_val)
                            else:
                                transformed = to_phonetic(src_val)
                            updates[new_field].append((transformed, pk_val))

                    with get_cursor(conn) as cursor:
                        for new_field, data in updates.items():
                            cursor.executemany(
                                f"UPDATE `{table_name}` SET `{new_field}`=%s WHERE `{effective_pk}`=%s",
                                data
                            )
                    conn.commit()

            created[table_name] = [f"{f.field_name}_{field_transform.value.lower()}" for f in flist]

        logger.info("✅ Finished create_table_fields.")
        return created

    def drop_table_fields(self, fields: list["TableField"], field_transform: FieldTransformType) -> dict[str, list[str]]:
        """
        Drop the derived columns referenced by TableField objects.
        Uses the provided field_transform to infer the column name:
        new_col = f"{field_name}_{field_transform.lower()}"

        Args:
            fields: List of TableField objects
            field_transform: Type of transform that was applied

        Returns: {table_name: [dropped_cols]}
        """
        # Group by table
        by_table: dict[str, list[TableField]] = {}
        for f in fields:
            by_table.setdefault(f.table_name, []).append(f)

        conn = self.connection_manager.connect()
        dropped: dict[str, list[str]] = {}

        for table, flist in by_table.items():
            # get current cols
            with get_cursor(conn) as c:
                c.execute(
                    "SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS "
                    "WHERE TABLE_SCHEMA=%s AND TABLE_NAME=%s",
                    (self.connection_manager.config.database, table)
                )
                cols = {r[0] for r in c.fetchall()}

            # build list of real cols to drop
            to_drop: list[str] = []
            for f in flist:
                col = f"__{f.field_name}_{field_transform.value.lower()}__"
                if col in cols:
                    to_drop.append(col)
                else:
                    logger.debug(f"Column '{col}' not found in '{table}', skipping.")

            if not to_drop:
                continue

            # one ALTER with multiple DROP COLUMN
            drops_sql = ", ".join(f"DROP COLUMN `{cname}`" for cname in to_drop)
            with get_cursor(conn) as c:
                c.execute(f"ALTER TABLE `{table}` {drops_sql}")
            conn.commit()

            dropped[table] = to_drop
            logger.info(f"Dropped columns on {table}: {', '.join(to_drop)}")

        return dropped

    def _get_or_create_table_pk(self, table_name: str) -> str:
        """
        Return a suitable PK/unique handle for batched updates.
        If none exists, create __tmp_pk__ as AUTO_INCREMENT PRIMARY KEY.
        """
        conn = self.connection_manager.connect()
        with get_cursor(conn) as c:
            c.execute("""
                SELECT COLUMN_NAME, COLUMN_KEY, EXTRA
                FROM INFORMATION_SCHEMA.COLUMNS
                WHERE TABLE_SCHEMA=%s AND TABLE_NAME=%s
            """, (self.connection_manager.config.database, table_name))
            rows = c.fetchall()

        cols = {r[0]: (r[1], r[2]) for r in rows}  # name -> (COLUMN_KEY, EXTRA)

        # Prefer real PK
        for name, (key, _) in cols.items():
            if key == "PRI":
                return name

        # Next best: unique, not null
        for name, (key, _) in cols.items():
            if key == "UNI":
                return name

        # Create throwaway
        tmp = "__tmp_pk__"
        if tmp not in cols:
            with get_cursor(conn) as c2:
                c2.execute(
                    f"ALTER TABLE `{table_name}` ADD COLUMN `{tmp}` BIGINT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY"
                )
            conn.commit()
        return tmp


    #########################################################
    # Table Dropping
    #########################################################
    def drop_table(self, table_name: str, skip_confirm: bool = False) -> None:
        """
        Delete a table from the database. Prompts for confirmation by default.
        Set skip_confirm=True to bypass confirmation.
        """
        clean_name = table_name.strip()
        if not skip_confirm:
            resp = input(
                f"Are you sure you want to delete table '{clean_name}'? This cannot be undone. [y/N]: "
            )
            if resp.strip().lower() not in ('y', 'yes'):
                logger.info(f"Deletion of table '{clean_name}' cancelled by user.")
                return

        conn = self.connection_manager.connect()
        current_db = self.connection_manager.config.database

        # Verify table exists
        with get_cursor(conn) as cursor:
            cursor.execute(
                "SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES "
                "WHERE TABLE_SCHEMA = %s AND TABLE_NAME = %s",
                (current_db, clean_name)
            )
            if not cursor.fetchone():
                logger.warning(f"Table '{clean_name}' does not exist; nothing to delete.")
                return

        # Drop the table
        with get_cursor(conn) as cursor:
            cursor.execute(f"DROP TABLE `{clean_name}`;")
        conn.commit()
        logger.info(f"Table '{clean_name}' deleted successfully.")


    #########################################################
    # Table Heuristics
    #########################################################
    def list_tables(self, database: str) -> list[str]:
        """
        List all tables in the specified database.
        """
        conn = self.connection_manager.connect()
        with get_cursor(conn) as cursor:
            cursor.execute(f"SHOW TABLES FROM {database};")
            return [row[0] for row in cursor.fetchall()]

    def get_table_size(self, table_name: str) -> int:
        """
        Get the size of a table in bytes.
        """
        conn = self.connection_manager.connect()
        with get_cursor(conn) as cursor:
            cursor.execute(f"SELECT SUM(DATA_LENGTH + INDEX_LENGTH) FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME = '{table_name}'")
            return cursor.fetchone()[0]




# canonmap/services/db_mysql/helpers/mysql_data_types.py

from enum import Enum

class MySQLType(str, Enum):
    # Numeric (exact value)
    TINYINT            = "TINYINT"
    SMALLINT           = "SMALLINT"
    MEDIUMINT          = "MEDIUMINT"
    INT                = "INT"
    INTEGER            = "INTEGER"
    BIGINT             = "BIGINT"
    DECIMAL            = "DECIMAL"
    NUMERIC            = "NUMERIC"
    FLOAT              = "FLOAT"
    DOUBLE             = "DOUBLE"
    REAL               = "REAL"
    BIT                = "BIT"
    BOOL               = "BOOL"
    BOOLEAN            = "BOOLEAN"
    SERIAL             = "SERIAL"

    # Date & Time
    DATE               = "DATE"
    DATETIME           = "DATETIME"
    TIMESTAMP          = "TIMESTAMP"
    TIME               = "TIME"
    YEAR               = "YEAR"

    # Character & Binary string
    CHAR               = "CHAR"
    VARCHAR            = "VARCHAR"
    BINARY             = "BINARY"
    VARBINARY          = "VARBINARY"
    TINYTEXT           = "TINYTEXT"
    TEXT               = "TEXT"
    MEDIUMTEXT         = "MEDIUMTEXT"
    LONGTEXT           = "LONGTEXT"
    TINYBLOB           = "TINYBLOB"
    BLOB               = "BLOB"
    MEDIUMBLOB         = "MEDIUMBLOB"
    LONGBLOB           = "LONGBLOB"
    ENUM               = "ENUM"
    SET                = "SET"

    # JSON
    JSON               = "JSON"

    # Spatial (GIS)
    GEOMETRY           = "GEOMETRY"
    POINT              = "POINT"
    LINESTRING         = "LINESTRING"
    POLYGON            = "POLYGON"
    MULTIPOINT         = "MULTIPOINT"
    MULTILINESTRING    = "MULTILINESTRING"
    MULTIPOLYGON       = "MULTIPOLYGON"
    GEOMETRYCOLLECTION = "GEOMETRYCOLLECTION"




import logging
import re
import unicodedata
from difflib import SequenceMatcher
from concurrent.futures import ThreadPoolExecutor, as_completed

from metaphone import doublemetaphone
try:
    import jellyfish
    _have_jaro = True
except ImportError:
    _have_jaro = False

try:
    import Levenshtein
    _have_lev = True
except ImportError:
    _have_lev = False


from canonmap.services.db_mysql.adapters.connection import ConnectionManager
from canonmap.services.db_mysql.schemas import EntityMappingRequest, EntityMappingResponse, SingleMappedEntity
from pydantic import BaseModel
from typing import List

logger = logging.getLogger(__name__)



class MatcherManager:
    def __init__(self, connection_manager: ConnectionManager):
        self.connection_manager = connection_manager

    def match(self, request: EntityMappingRequest, weights=None) -> EntityMappingResponse:
        def _normalize(s: str) -> str:
            s = unicodedata.normalize("NFKD", s).encode("ascii", "ignore").decode("ascii")
            s = re.sub(r"[^\w\s]", " ", s)
            return re.sub(r"\s+", " ", s).strip().lower()

        def _trigram_similarity(a: str, b: str) -> float:
            def grams(s):
                return {s[i:i+3] for i in range(len(s)-2)}
            A, B = grams(a), grams(b)
            if not A or not B:
                return 0.0
            return len(A & B) / len(A | B)

        def _block_by_phonetic(conn, entity_name: str, table_name: str, field_name: str) -> set:
            p1, p2 = doublemetaphone(entity_name)
            search_code = p1 or p2
            if not search_code:
                return set()
            sql = f"""SELECT DISTINCT `{field_name}` AS name
                    FROM `{table_name}`
                    WHERE `__{field_name}_phonetic__` = %s"""
            with conn.cursor() as cur:
                cur.execute(sql, (search_code,))
                return {r[0] for r in cur.fetchall()}

        def _block_by_soundex(conn, entity_name: str, table_name: str, field_name: str) -> set:
            sql = f"""SELECT DISTINCT `{field_name}` AS name
                    FROM `{table_name}`
                    WHERE `__{field_name}_soundex__` = SOUNDEX(%s)"""
            with conn.cursor() as cur:
                cur.execute(sql, (entity_name,))
                return {r[0] for r in cur.fetchall()}

        def _block_by_initialism(conn, entity_name: str, table_name: str, field_name: str) -> set:
            if not entity_name:
                return set()
            entity_clean = entity_name.strip().upper()
            if (entity_clean.isalpha() and 
                len(entity_clean) <= 6 and 
                len(entity_clean) >= 2 and
                ' ' not in entity_clean):
                search_initialism = entity_clean
            else:
                parts = re.findall(r"[A-Za-z]+", entity_name)
                search_initialism = "".join(p[0].upper() for p in parts) if parts else None
            if not search_initialism:
                return set()
            sql = f"""SELECT DISTINCT `{field_name}` AS name
                    FROM `{table_name}`
                    WHERE `__{field_name}_initialism__` = %s"""
            with conn.cursor() as cur:
                cur.execute(sql, (search_initialism,))
                return {r[0] for r in cur.fetchall()}

        def _block_by_exact_match(conn, entity_name: str, table_name: str, field_name: str) -> set:
            if not entity_name:
                return set()
            search_term = entity_name.strip().lower()
            sql = f"""SELECT DISTINCT `{field_name}` AS name
                    FROM `{table_name}`
                    WHERE LOWER(TRIM(`{field_name}`)) LIKE %s"""
            with conn.cursor() as cur:
                cur.execute(sql, (f"%{search_term}%",))
                return {r[0] for r in cur.fetchall()}

        default_weights = {
            "exact": 6.0,
            "levenshtein": 1.0,
            "jaro": 1.2,
            "token": 2.0,
            "trigram": 1.0,
            "phonetic": 1.0,
            "initialism": 0.5,
            "multi_bonus": 1.0,
        }
        weights = weights or default_weights

        normalized_entity_name = _normalize(request.entity_name)
        candidates = set()
        conn = self.connection_manager.conn or self.connection_manager.connect()
        # for field in request.select_fields:
        phonetic_candidates = _block_by_phonetic(conn, normalized_entity_name, request.select_field.table_name, request.select_field.field_name)
        soundex_candidates = _block_by_soundex(conn, normalized_entity_name, request.select_field.table_name, request.select_field.field_name)
        initialism_candidates = _block_by_initialism(conn, normalized_entity_name, request.select_field.table_name, request.select_field.field_name)
        exact_candidates = _block_by_exact_match(conn, normalized_entity_name, request.select_field.table_name, request.select_field.field_name)
        
        candidates = candidates.union(phonetic_candidates)
        candidates = candidates.union(soundex_candidates)
        candidates = candidates.union(initialism_candidates)
        candidates = candidates.union(exact_candidates)
        
        logger.debug(f"TableField {request.select_field.field_name}: phonetic={len(phonetic_candidates)}, soundex={len(soundex_candidates)}, initialism={len(initialism_candidates)}, exact={len(exact_candidates)}")
        
        logger.debug(f"Total candidates found: {len(candidates)}")
        
        # Always expand the search to get a good candidate pool, regardless of top_n
        # This ensures consistent rankings regardless of how many results are requested
        min_candidates = max(50, request.top_n * 3)  # Get at least 50 candidates or 3x top_n
        if len(candidates) < min_candidates:
            logger.debug(f"Only found {len(candidates)} candidates, expanding search to get at least {min_candidates} candidates")
            
            # Add a broader search to get more candidates
            def _get_more_candidates(conn, entity_name: str, table_name: str, field_name: str, min_candidates: int) -> set:
                additional_candidates = set()
                
                # Try partial name matching
                if len(entity_name.split()) > 1:
                    first_name = entity_name.split()[0]
                    last_name = entity_name.split()[-1]
                    
                    # Search for names containing first or last name
                    sql = f"""SELECT DISTINCT `{field_name}` AS name
                            FROM `{table_name}`
                            WHERE LOWER(TRIM(`{field_name}`)) LIKE %s OR LOWER(TRIM(`{field_name}`)) LIKE %s
                            LIMIT %s"""
                    with conn.cursor() as cur:
                        cur.execute(sql, (f"%{first_name}%", f"%{last_name}%", min_candidates * 2))
                        additional_candidates.update(r[0] for r in cur.fetchall())
                

                
                # If still not enough, get some random candidates
                if len(additional_candidates) < min_candidates:
                    sql = f"""SELECT DISTINCT `{field_name}` AS name
                            FROM `{table_name}`
                            WHERE `{field_name}` IS NOT NULL AND `{field_name}` != ''
                            ORDER BY RAND()
                            LIMIT %s"""
                    with conn.cursor() as cur:
                        cur.execute(sql, (min_candidates,))
                        additional_candidates.update(r[0] for r in cur.fetchall())
                
                return additional_candidates
            
            additional = _get_more_candidates(conn, normalized_entity_name, request.select_field.table_name, request.select_field.field_name, min_candidates - len(candidates))
            candidates = candidates.union(additional)
            logger.debug(f"Added {len(additional)} additional candidates from broader search")
        
        logger.debug(f"Final total candidates: {len(candidates)}")

        def _score_candidate(normalized_entity_name: str, candidate_name: str) -> dict:
            cand_norm = _normalize(candidate_name)
            tokens = normalized_entity_name.split()
            first, last = tokens[0], tokens[-1] if tokens else ("", "")
            # exact
            exact = 1.0 if cand_norm == normalized_entity_name else 0.0
            # Levenshtein
            if _have_lev:
                lev_full = Levenshtein.ratio(normalized_entity_name, cand_norm)
                lev_last = Levenshtein.ratio(last, _normalize(candidate_name.split()[-1])) if last else 0.0
            else:
                lev_full = SequenceMatcher(None, normalized_entity_name, cand_norm).ratio()
                lev_last = SequenceMatcher(None, last, _normalize(candidate_name.split()[-1])).ratio() if last else 0.0
            levenshtein = 0.3 * lev_full + 0.7 * lev_last
            # Jaro–Winkler
            if _have_jaro:
                jaro = jellyfish.jaro_winkler_similarity(normalized_entity_name, cand_norm)
            else:
                jaro = levenshtein
            # Token overlap (first vs last)
            tok_first = float(first in cand_norm)
            tok_last = float(last in cand_norm)
            token = 0.3 * tok_first + 0.7 * tok_last
            # Trigram
            tri = _trigram_similarity(normalized_entity_name, cand_norm)
            # Phonetic - recompute from candidate name
            p1, p2 = doublemetaphone(cand_norm)
            last_phonetic = doublemetaphone(last)[0] if last else ""
            phon = float(last_phonetic in (p1, p2)) if last_phonetic else 0.0
            # Initialism - recompute from candidate name
            init = "".join(tok[0] for tok in cand_norm.split() if tok)
            query_init = "".join(tok[0] for tok in normalized_entity_name.split() if tok)
            init_score = float(init == query_init) if query_init else 0.0
            
            return {
                "exact": exact,
                "levenshtein": levenshtein,
                "jaro": jaro,
                "token": token,
                "trigram": tri,
                "phonetic": phon,
                "initialism": init_score,
            }

        # Score in parallel
        signatures = []
        with ThreadPoolExecutor() as ex:
            futures = {ex.submit(_score_candidate, normalized_entity_name, candidate_name): candidate_name for candidate_name in candidates}
            for future in as_completed(futures):
                candidate_name = futures[future]
                signature = future.result()
                signatures.append((candidate_name, signature))

        # Combine + rank
        ranked = []
        for candidate_name, signature in signatures:
            total = sum(signature[k] * weights[k] for k in signature)
            multi = sum(1 for k in ("levenshtein","token","phonetic","initialism") if signature[k] > 0)
            total += max(0, multi - 1) * weights["multi_bonus"]
            ranked.append((candidate_name, total))

        ranked.sort(key=lambda x: x[1], reverse=True)
        
        # Return top_n results
        return EntityMappingResponse(results=[
            SingleMappedEntity(
                raw_entity=request.entity_name, 
                canonical_entity=candidate_name, 
                canonical_table_name=request.select_field.table_name, 
                canonical_field_name=request.select_field.field_name, 
                score=score) 
            for candidate_name, score in ranked[:request.top_n]
        ])
    



# canonmap/services/db_mysql/managers/user.py

import logging
from typing import Optional

from canonmap.services.db_mysql.adapters.connection import ConnectionManager
from canonmap.services.db_mysql.adapters.cursor import get_cursor

logger = logging.getLogger(__name__)


class UserManager:
    def __init__(self, connection_manager: ConnectionManager):
        self.connection_manager = connection_manager

    def create_user(
        self,
        username: str,
        password: str,
        host: str = '%',
        privileges: Optional[str] = None,
        on_database: Optional[str] = None
    ) -> None:
        conn = self.connection_manager.connect()
        with get_cursor(conn) as cursor:
            cursor.execute(
                f"CREATE USER IF NOT EXISTS `{username}`@`{host}` IDENTIFIED BY %s",
                (password,)
            )
            if privileges:
                target = f"`{on_database}`.*" if on_database else "*.*"
                cursor.execute(
                    f"GRANT {privileges} ON {target} TO `{username}`@`{host}`"
                )
            cursor.execute("FLUSH PRIVILEGES")
        conn.commit()

        logger.info(
            f"User '{username}'@'{host}' has been created"
            + (f" with privileges '{privileges}' on {on_database}" if privileges else "")
        )

    def delete_user(self, username: str, host: str = '%') -> None:
        conn = self.connection_manager.connect()
        with get_cursor(conn) as cursor:
            cursor.execute(f"DROP USER IF EXISTS `{username}`@`{host}`")
            cursor.execute("FLUSH PRIVILEGES")
        conn.commit()

        logger.info(f"User '{username}'@'{host}' has been deleted")


# canonmap/utils/logger.py

import logging

from rich.logging import RichHandler

def configure_logging(level: str = "INFO") -> None:
    """
    Call once, very early in your program.
    The force=True parameter makes sure basicConfig replaces existing handlers.
    """
    logging.basicConfig(
        level=level,
        format="%(asctime)s %(name)s %(levelname)s: %(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(show_time=False, markup=True)],
        force=True,
    )


def configure_logging_decorator(level: str = "INFO"):
    """
    Decorator version of configure_logging.
    Usage: @configure_logging_decorator("INFO")
    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            configure_logging(level)
            return func(*args, **kwargs)
        return wrapper
    return decorator


from setuptools import setup, find_packages

setup(
    name="canonmap",
    version="0.3.39",
    packages=find_packages(),
    install_requires=[
        "python-dotenv",
        "pydantic",
        "pandas",
        "numpy",
        "rich",
        "mysql-connector-python",
        "metaphone",
        "python-Levenshtein",
        "jellyfish",
    ],
    extras_require={},
    author="Vince Berry",
    author_email="vince.berry@gmail.com",
    description="CanonMap - A Python library for entity canonicalization and mapping with enhanced configuration and response models",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/vinceberry/canonmap",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.8",
    include_package_data=True,
    entry_points={},
)


# canonmap/services/db_mysql/helpers/cloud_sql.py

import os
import logging
from typing import Optional

from canonmap.services.db_mysql.config import CanonMapMySQLConfig
from canonmap.services.db_mysql.core import CanonMapMySQLClient

logger = logging.getLogger(__name__)

def create_cloud_sql_client(
    project_id: Optional[str] = None,
    region: Optional[str] = None,
    instance_name: Optional[str] = None,
    database_name: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
) -> CanonMapMySQLClient:
    """
    Create a CanonMapMySQLClient configured for Google Cloud SQL.
    
    Args:
        project_id: GCP project ID (defaults to GOOGLE_CLOUD_PROJECT env var)
        region: Cloud SQL region (defaults to DB_REGION env var)
        instance_name: Cloud SQL instance name (defaults to DB_INSTANCE env var)
        database_name: Database name (defaults to DB_NAME env var)
        user: MySQL username (defaults to DB_USER env var)
        password: MySQL password (defaults to DB_PASSWORD env var)
    
    Returns:
        CanonMapMySQLClient configured for Cloud SQL
    """
    # Get values from environment if not provided
    project_id = project_id or os.getenv("GOOGLE_CLOUD_PROJECT")
    region = region or os.getenv("DB_REGION")
    instance_name = instance_name or os.getenv("DB_INSTANCE")
    database_name = database_name or os.getenv("DB_NAME")
    user = user or os.getenv("DB_USER")
    password = password or os.getenv("DB_PASSWORD")
    
    if not all([project_id, region, instance_name, database_name, user, password]):
        raise ValueError(
            "Missing required Cloud SQL configuration. Please provide all parameters "
            "or set the following environment variables:\n"
            "- GOOGLE_CLOUD_PROJECT (or project_id parameter)\n"
            "- DB_REGION (or region parameter)\n"
            "- DB_INSTANCE (or instance_name parameter)\n"
            "- DB_NAME (or database_name parameter)\n"
            "- DB_USER (or user parameter)\n"
            "- DB_PASSWORD (or password parameter)"
        )
    
    # Construct Unix socket path for Cloud SQL
    unix_socket = f"/cloudsql/{project_id}:{region}:{instance_name}"
    
    config = CanonMapMySQLConfig(
        host="localhost",  # Not used with Unix socket
        user=user,
        password=password,
        database=database_name,
        unix_socket=unix_socket,
    )
    
    logger.info(f"Creating Cloud SQL client for instance: {project_id}:{region}:{instance_name}")
    return CanonMapMySQLClient(config)

def create_client_from_env() -> CanonMapMySQLClient:
    """
    Create a CanonMapMySQLClient using environment variables.
    
    This function will automatically detect whether to use Cloud SQL (Unix socket)
    or regular MySQL (TCP) based on the presence of DB_UNIX_SOCKET environment variable.
    
    Environment variables:
    - For Cloud SQL: DB_UNIX_SOCKET, DB_USER, DB_PASSWORD, DB_NAME
    - For regular MySQL: DB_HOST, DB_PORT, DB_USER, DB_PASSWORD, DB_NAME
    """
    config = CanonMapMySQLConfig.from_environment()
    logger.info("Creating MySQL client from environment configuration")
    return CanonMapMySQLClient(config) 